#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

#define WARP_SIZE 32

// Each block computes one dot product for (batch_idx, out_idx)
// ensuring memory coalescing by having threads in the warp load consecutive elements
// from the input row of x and the corresponding row of linear_weight.

__global__ void my_kernel(
    const float* x,
    const float* linear_weight,
    const float* linear_bias,
    const float* constant,
    float* y,
    int batch_size,
    int in_features,
    int out_features) {
    // Map each block to one output element for one batch element
    int index = blockIdx.x;  // index over (batch * out_features)
    int batch_idx = index / out_features;
    int out_idx = index % out_features;

    float sum = 0.0f;
    // Each thread in the block (warp) processes a portion of the dot product
    for (int i = threadIdx.x; i < in_features; i += blockDim.x) {
        // Coalesced read: threads in warp read consecutive elements
        float a = x[batch_idx * in_features + i];
        float b = linear_weight[out_idx * in_features + i];
        sum += a * b;
    }

    // Warp-level reduction using __shfl_down_sync
    for (int offset = WARP_SIZE / 2; offset > 0; offset /= 2) {
        sum += __shfl_down_sync(0xffffffff, sum, offset);
    }

    if (threadIdx.x == 0) {
        float result = sum + linear_bias[out_idx];
        float c = *constant;
        result = (result > c) ? c : result;
        y[batch_idx * out_features + out_idx] = result - c;
    }
}

// Pybind11 binding

torch::Tensor forward(
    torch::Tensor x,
    torch::Tensor linear_weight,
    torch::Tensor linear_bias,
    torch::Tensor constant) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(linear_weight.is_cuda(), "linear_weight must be a CUDA tensor");
    TORCH_CHECK(linear_bias.is_cuda(), "linear_bias must be a CUDA tensor");
    TORCH_CHECK(constant.is_cuda(), "constant must be a CUDA tensor");

    int batch_size = x.size(0);
    int in_features = x.size(1);
    int out_features = linear_weight.size(0);

    auto y = torch::zeros({batch_size, out_features}, x.options());

    const float* x_ptr = x.data_ptr<float>();
    const float* weight_ptr = linear_weight.data_ptr<float>();
    const float* bias_ptr = linear_bias.data_ptr<float>();
    const float* constant_ptr = constant.data_ptr<float>();
    float* y_ptr = y.data_ptr<float>();

    // Launch one block per (batch element, out_feature) pair.
    int total_blocks = batch_size * out_features;
    int threads = WARP_SIZE;  // 32 threads per block to form a full warp

    my_kernel<<<total_blocks, threads>>>(
        x_ptr,
        weight_ptr,
        bias_ptr,
        constant_ptr,
        y_ptr,
        batch_size,
        in_features,
        out_features);

    return y;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "CUDA forward function");
}
